/*-
 * Copyright (C) 2009-2010 Semihalf, Grzegorz Bernacki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Generic NAND driver */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/rman.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/time.h>
#include <sys/malloc.h>

#include <dev/nand/nand.h>
#include <dev/nand/nandbus.h>
#include "nfc_if.h"
#include "nand_if.h"

#define DEBUG
#undef DEBUG
#ifdef DEBUG
#define debug(fmt, args...) do { \
    printf("NAND_CHIP:" fmt "\n", ##args); } while (0)
#else
#define debug(fmt, args...)
#endif

static int onfi_nand_probe(device_t dev);
static int large_nand_probe(device_t dev);
static int small_nand_probe(device_t dev);
static int generic_nand_attach(device_t dev);
static int generic_nand_detach(device_t dev);

static int generic_erase_block(device_t nand, uint32_t block);
static int generic_erase_block_intlv(device_t nand, uint32_t block);
static int generic_read_page (device_t dev, uint32_t page, void* buf,
    uint32_t len, uint32_t offset);
static int generic_read_oob(device_t nand, uint32_t page, void* buf,
    uint32_t len, uint32_t offset);
static int generic_program_page(device_t nand, uint32_t page, void *buf,
    uint32_t len, uint32_t offset);
static int generic_program_page_intlv(device_t nand, uint32_t page, void *buf,
    uint32_t len, uint32_t offset);
static int generic_program_oob(device_t nand, uint32_t page, void* buf,
    uint32_t len, uint32_t offset);

static int small_read_page(device_t nand, uint32_t page, void *buf,
    uint32_t len, uint32_t offset);
static int small_read_oob(device_t nand, uint32_t page, void *buf,
    uint32_t len, uint32_t offset);
static int small_program_page(device_t nand, uint32_t page, void* buf,
    uint32_t len, uint32_t offset);
static int small_program_oob(device_t nand, uint32_t page, void* buf,
    uint32_t len, uint32_t offset);

static int onfi_is_blk_bad(device_t device, uint32_t block, uint8_t *bad);
static int onfi_read_parameter(struct nand_chip *chip,
    struct onfi_params *params);

static device_method_t onand_methods[] = {
	/* device interface */
	DEVMETHOD(device_probe,			onfi_nand_probe),
	DEVMETHOD(device_attach,		generic_nand_attach),
	DEVMETHOD(device_detach,		generic_nand_detach),

	DEVMETHOD(nand_read_page,		generic_read_page),
	DEVMETHOD(nand_program_page,		generic_program_page),
	DEVMETHOD(nand_program_page_intlv,	generic_program_page_intlv),
	DEVMETHOD(nand_read_oob,		generic_read_oob),
	DEVMETHOD(nand_program_oob,		generic_program_oob),
	DEVMETHOD(nand_erase_block,		generic_erase_block),
	DEVMETHOD(nand_erase_block_intlv,	generic_erase_block_intlv),

	DEVMETHOD(nand_is_blk_bad,		onfi_is_blk_bad),
	{ 0, 0 }
};

static device_method_t lnand_methods[] = {
	/* device interface */
	DEVMETHOD(device_probe,		large_nand_probe),
	DEVMETHOD(device_attach,	generic_nand_attach),
	DEVMETHOD(device_detach,	generic_nand_detach),

	DEVMETHOD(nand_read_page,	generic_read_page),
	DEVMETHOD(nand_program_page,	generic_program_page),
	DEVMETHOD(nand_read_oob,	generic_read_oob),
	DEVMETHOD(nand_program_oob,	generic_program_oob),
	DEVMETHOD(nand_erase_block,	generic_erase_block),
	{ 0, 0 }
};

static device_method_t snand_methods[] = {
	/* device interface */
	DEVMETHOD(device_probe,		small_nand_probe),
	DEVMETHOD(device_attach,	generic_nand_attach),
	DEVMETHOD(device_detach,	generic_nand_detach),

	DEVMETHOD(nand_read_page,	small_read_page),
	DEVMETHOD(nand_program_page,	small_program_page),
	DEVMETHOD(nand_read_oob,	small_read_oob),
	DEVMETHOD(nand_program_oob,	small_program_oob),
	DEVMETHOD(nand_erase_block,	generic_erase_block),
	{ 0, 0 }
};

devclass_t onand_devclass;
devclass_t lnand_devclass;
devclass_t snand_devclass;

driver_t onand_driver = {
	"onand",
	onand_methods,
	sizeof(struct nand_chip)
};

driver_t lnand_driver = {
	"lnand",
	lnand_methods,
	sizeof(struct nand_chip)
};

driver_t snand_driver = {
	"snand",
	snand_methods,
	sizeof(struct nand_chip)
};

DRIVER_MODULE(onand, nandbus, onand_driver, onand_devclass, 0, 0);
DRIVER_MODULE(lnand, nandbus, lnand_driver, lnand_devclass, 0, 0);
DRIVER_MODULE(snand, nandbus, snand_driver, snand_devclass, 0, 0);

static int
onfi_nand_probe(device_t dev)
{
	struct nandbus_ivar *ivar;

	ivar = device_get_ivars(dev);
	if (ivar && ivar->is_onfi) {
		device_set_desc(dev, "ONFI compliant NAND");
		return (BUS_PROBE_DEFAULT);
	}

	return (ENODEV);
}

static int
large_nand_probe(device_t dev)
{
	struct nandbus_ivar *ivar;

	ivar = device_get_ivars(dev);
	if (ivar && !ivar->is_onfi &&
	    (ivar->man_id == NAND_MAN_SAMSUNG) && (ivar->dev_id == 0xd3)) {
		device_set_desc(dev, "Large Samsung NAND");
		return (BUS_PROBE_DEFAULT);
	}

	return (ENODEV);
}

static int
small_nand_probe(device_t dev)
{
	struct nandbus_ivar *ivar;

	ivar = device_get_ivars(dev);
	if (ivar && !ivar->is_onfi &&
	    (ivar->man_id == NAND_MAN_SAMSUNG) && (ivar->dev_id == 0x75)) {
		device_set_desc(dev, "Small Samsung NAND");
		return (BUS_PROBE_DEFAULT);
	}

	return (ENODEV);
}

static int
generic_nand_attach(device_t dev)
{
	struct nand_chip *chip;
	struct nand_params *chip_params;
	struct nandbus_ivar *ivar;
	struct onfi_params *onfi_params;
	device_t nandbus, nfc;
	int err;

	chip = device_get_softc(dev);
	chip->dev = dev;

	ivar = device_get_ivars(dev);
	chip->id.manufacturer_id = ivar->man_id;
	chip->id.device_id = ivar->dev_id;
	chip->num = ivar->cs;

	/* TODO remove when HW ECC supported */
	nandbus = device_get_parent(dev);
	nfc = device_get_parent(nandbus);
	chip->nand = device_get_softc(nfc);

	if (ivar->is_onfi) {
		onfi_params = malloc(sizeof(struct onfi_params),
			M_DEVBUF, M_WAITOK | M_ZERO);
		if (onfi_params == NULL)
			return (ENXIO);

		if (onfi_read_parameter(chip, onfi_params)) {
			debug("Could not read parameter page!\n");
			free(onfi_params, M_DEVBUF);
			return (ENXIO);
		}

		nand_onfi_set_params(chip, onfi_params);
		free(onfi_params, M_DEVBUF);

	} else {
		chip_params = nand_get_params(&chip->id);
		if (chip_params == NULL) {
			debug("Chip description not found!"
				"(manuf: 0x%0x, chipid: 0x%0x)\n",
				chip->id.manufacturer_id,
				chip->id.device_id);
			return (EINVAL);
		}

		nand_set_params(chip, chip_params);
	}

	err = nand_init_stat(chip);
	if (err) {
		generic_nand_detach(dev);
		return (err);
	}

#if 0
	err = nand_init_bbt(chip);
	if (err) {
		generic_nand_detach(dev);
		return (err);
	}
#endif

	err = nand_make_dev(chip);
	if (err) {
		generic_nand_detach(dev);
		return (err);
	}

	err = create_geom_disk(chip);
	if (err) {
		generic_nand_detach(dev);
		return (err);
	}

	return (0);
}

static int
generic_nand_detach(device_t dev)
{
	struct nand_chip *chip;

	chip = device_get_softc(dev);

	nand_destroy_bbt(chip);
	destroy_geom_disk(chip);
	nand_destroy_dev(chip);
	nand_destroy_stat(chip);

	return (0);
}

static int
can_write(device_t nandbus)
{
	uint8_t status;

	if (nandbus_wait_ready(nandbus, &status))
		return (0);

	if (!(status & NAND_STATUS_WP)) {
		debug("Chip is write-protected");
		return (0);
	}

	return (1);
}

static int
check_fail(device_t nandbus)
{
	uint8_t status;

	nandbus_wait_ready(nandbus, &status);
	if (status & NAND_STATUS_FAIL) {
		debug("Status failed %x", status);
		return (ENXIO);
	}

	return (0);
}

static int
onfi_read_parameter(struct nand_chip *chip, struct onfi_params *params)
{
	device_t nandbus;

	debug("read parameter");

	nandbus = device_get_parent(chip->dev);

	nandbus_select_cs(nandbus, chip->num);

	if (nandbus_send_command(nandbus, NAND_CMD_READ_PARAMETER))
		return (ENXIO);

	if (nandbus_send_address(nandbus, -1, -1, PAGE_PARAMETER_DEF))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	nandbus_read_buffer(nandbus, params, PAGE_PARAM_LENGTH);

	/* TODO */
	/* Check for signature */
	/* Check CRC */
	/* Use redundant page if necessary */

	return (0);
}

static int
send_read_page(device_t nandbus, uint8_t start_command, uint8_t end_command,
    uint32_t row, uint32_t column)
{
	if (nandbus_send_command(nandbus, start_command))
		return (ENXIO);

	if (nandbus_send_address(nandbus, row, column, -1))
		return (ENXIO);

	if (nandbus_send_command(nandbus, end_command))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	return (0);
}

static int
generic_read_page(device_t nand, uint32_t page, void *buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug("%p raw read page %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (send_read_page(nandbus, NAND_CMD_READ, NAND_CMD_READ_END, row,
		    offset))
		return (ENXIO);

	DELAY(chip->t_r);
	if (check_fail(nandbus))
		return (ENXIO);

	nandbus_read_buffer(nandbus, buf, len);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_read++;

	return (0);
}

static int
generic_read_oob(device_t nand, uint32_t page, void* buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	device_t nandbus;
	uint32_t row;

	debug("%p raw read oob %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	offset += chip->chip_geom.page_size;

	if (send_read_page(nandbus, NAND_CMD_READ, NAND_CMD_READ_END, row,
		    offset))
		return (ENXIO);

	DELAY(chip->t_r);

	if (check_fail(nandbus))
		return (ENXIO);

	nandbus_read_buffer(nandbus, buf, len);

	return (0);
}

static int
send_start_program_page(device_t nandbus, uint32_t row, uint32_t column)
{
	if (nandbus_send_command(nandbus, NAND_CMD_PROG))
		return (ENXIO);

	if (nandbus_send_address(nandbus, row, column, -1))
		return (ENXIO);

	return (0);
}

static int
send_end_program_page(device_t nandbus, uint8_t end_command)
{

	if (nandbus_send_command(nandbus, end_command))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	return (0);
}

static int
generic_program_page(device_t nand, uint32_t page, void *buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug("%p raw prog page %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (!can_write(nandbus))
		return (ENXIO);

	if (send_start_program_page(nandbus, row, offset))
		return (ENXIO);

	nandbus_write_buffer(nandbus, buf, len);

	if (send_end_program_page(nandbus, NAND_CMD_PROG_END))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_written++;

	return (0);
}

static int
generic_program_page_intlv(device_t nand, uint32_t page, void *buf,
    uint32_t len, uint32_t offset)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug("%p raw prog page %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (!can_write(nandbus))
		return (ENXIO);

	if (send_start_program_page(nandbus, row, offset))
		return (ENXIO);

	nandbus_write_buffer(nandbus, buf, len);

	if (send_end_program_page(nandbus, NAND_CMD_PROG_INTLV))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_written++;

	return (0);
}

static int
generic_program_oob(device_t nand, uint32_t page, void* buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	device_t nandbus;
	uint32_t row;

	debug("%p raw prog oob %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);
	offset += chip->chip_geom.page_size;

	if (!can_write(nandbus))
		return (ENXIO);

	if (send_start_program_page(nandbus, row, offset))
		return (ENXIO);

	nandbus_write_buffer(nandbus, buf, len);

	if (send_end_program_page(nandbus, NAND_CMD_PROG_END))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	return (0);
}

static int
send_erase_block(device_t nandbus, uint32_t row, uint8_t second_command)
{

	if (nandbus_send_command(nandbus, NAND_CMD_ERASE))
		return (ENXIO);

	if (nandbus_send_address(nandbus, row, -1, -1))
		return (ENXIO);

	if (nandbus_send_command(nandbus, second_command))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	return (0);
}

static int
generic_erase_block(device_t nand, uint32_t block)
{
	struct block_stat *blk_stat;
	struct nand_chip *chip;
	device_t nandbus;
	int row;

	debug("%p erase block  %x", nand, block);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (block >= (chip->chip_geom.blks_per_lun * chip->chip_geom.luns))
		return (ENXIO);

	row = (block << chip->chip_geom.blk_shift) &
	    chip->chip_geom.blk_mask;

	debug("%p erase block  row %x", nand, row);

	if (!can_write(nandbus))
		return (ENXIO);

	send_erase_block(nandbus, row, NAND_CMD_ERASE_END);

	DELAY(chip->t_bers);

	if (check_fail(nandbus))
		return (ENXIO);

	blk_stat = &(chip->blk_stat[block]);
	blk_stat->block_erased++;

	return (0);
}

static int
generic_erase_block_intlv(device_t nand, uint32_t block)
{
	struct block_stat *blk_stat;
	struct nand_chip *chip;
	device_t nandbus;
	int row;

	debug("%p erase block  %x", nand, block);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (block >= (chip->chip_geom.blks_per_lun * chip->chip_geom.luns))
		return (ENXIO);

	row = (block << chip->chip_geom.blk_shift) &
	    chip->chip_geom.blk_mask;

	if (!can_write(nandbus))
		return (ENXIO);

	send_erase_block(nandbus, row, NAND_CMD_ERASE_INTLV);

	DELAY(chip->t_bers);

	if (check_fail(nandbus))
		return (ENXIO);

	blk_stat = &(chip->blk_stat[block]);
	blk_stat->block_erased++;

	return (0);

}

static int
onfi_is_blk_bad(device_t device, uint32_t block_number, uint8_t *bad)
{
	struct nand_chip *chip;
	int page_number, i, j, err;
	uint8_t *oob;

	chip = device_get_softc(device);

	oob = malloc(chip->chip_geom.oob_size, M_DEVBUF, M_WAITOK);
	if (!oob) {
		device_printf(device, "%s: cannot allocate oob\n", __func__);
		return (ENOMEM);
	}

	page_number = block_number * chip->chip_geom.pgs_per_blk;
	*bad = 0;
	/* check oob of first and last page */
	for (i = 0; i < 2; i++, page_number+= chip->chip_geom.pgs_per_blk - 1) {
		err = generic_read_oob(device, page_number, oob,
		    chip->chip_geom.oob_size, 0);
		if (err) {
			device_printf(device, "%s: cannot allocate oob\n",
			    __func__);
			free(oob, M_DEVBUF);
			return (ENOMEM);
		}

		for (j = 0; j < chip->chip_geom.oob_size; j++) {
			if (!oob[j]) {
				*bad = 1;
				free(oob, M_DEVBUF);
				return (0);
			}
		}
	}

	free(oob, M_DEVBUF);

	return (0);
}

static int
send_small_read_page(device_t nandbus, uint8_t start_command,
    uint32_t row, uint32_t column)
{
	if (nandbus_send_command(nandbus, start_command))
		return (ENXIO);

	if (nandbus_send_address(nandbus, row, column, -1))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	return (0);
}


static int
small_read_page(device_t nand, uint32_t page, void *buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug("%p small read page %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (offset < 256) {
		if (send_small_read_page(nandbus, NAND_CMD_SMALLA, row, offset))
			return (ENXIO);
	} else {
		if (send_small_read_page(nandbus, NAND_CMD_SMALLB, row, offset))
			return (ENXIO);
	}

	DELAY(chip->t_r);

	nandbus_read_buffer(nandbus, buf, len);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_read++;

	return (0);
}

static int
small_read_oob(device_t nand, uint32_t page, void *buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug("%p small read oob %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (send_small_read_page(nandbus, NAND_CMD_SMALLOOB, row, offset))
		return (ENXIO);

	DELAY(chip->t_r);

	nandbus_read_buffer(nandbus, buf, len);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_read++;

	return (0);
}

static int
small_program_page(device_t nand, uint32_t page, void* buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	device_t nandbus;
	uint32_t row;

	debug("%p small prog page %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (!can_write(nandbus))
		return (ENXIO);

	if (offset < 256) {
		if (nandbus_send_command(nandbus, NAND_CMD_SMALLA))
			return (ENXIO);
	} else {
		if (nandbus_send_command(nandbus, NAND_CMD_SMALLB))
			return (ENXIO);
	}

	if (send_start_program_page(nandbus, row, offset))
		return (ENXIO);

	nandbus_write_buffer(nandbus, buf, len);

	if (send_end_program_page(nandbus, NAND_CMD_PROG_END))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	return (0);
}

static int
small_program_oob(device_t nand, uint32_t page, void* buf, uint32_t len,
    uint32_t offset)
{
	struct nand_chip *chip;
	device_t nandbus;
	uint32_t row;

	debug("%p small prog oob %x[%x] at %x", nand, page, len, offset);
	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (!can_write(nandbus))
		return (ENXIO);

	if (nandbus_send_command(nandbus, NAND_CMD_SMALLOOB))
		return (ENXIO);

	if (send_start_program_page(nandbus, row, offset))
		return (ENXIO);

	nandbus_write_buffer(nandbus, buf, len);

	if (send_end_program_page(nandbus, NAND_CMD_PROG_END))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	return (0);
}

#if 0
int
nand_chng_read_col(device_t nand, uint32_t col, void *buf, size_t len)
{
	struct nand_chip *chip;
	device_t nandbus;

	chip = device_get_softc(nand);
	nandbus = device_get_parent(nand);

	if (nandbus_send_command(nandbus, NAND_CMD_CHNG_READ_COL))
		return (ENXIO);

	if (nandbus_send_address(nandbus, -1, col, -1))
		return (ENXIO);

	if (nandbus_send_command(nandbus, NAND_CMD_CHNG_READ_COL_END))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	if (buf != NULL && len > 0)
		nandbus_read_buffer(nandbus, buf, len);

	return (0);
}

int
nand_chng_write_col(device_t dev, uint32_t col, void *buf,
    size_t len)
{
	struct nand_chip *chip;
	device_t nandbus;

	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nandbus_send_command(nandbus, NAND_CMD_CHNG_WRITE_COL))
		return (ENXIO);

	if (nandbus_send_address(nandbus, -1, col, -1))
		return (ENXIO);

	if (buf != NULL && len > 0)
		nandbus_write_buffer(nandbus, buf, len);

	if (nandbus_send_command(nandbus, NAND_CMD_CHNG_READ_COL_END))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	return (0);
}

int
nand_copyback_read(device_t dev, uint32_t page, uint32_t col,
    void *buf, size_t len)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug(" raw read page %x[%x] at %x", page, col, len);
	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (send_read_page(nandbus, NAND_CMD_READ, NAND_CMD_READ_CPBK, row, 0))
		return (ENXIO);

	DELAY(chip->t_r);
	if (check_fail(nandbus))
		return (ENXIO);

	if (buf != NULL && len > 0)
		nandbus_read_buffer(nandbus, buf, len);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_read++;

	return (0);
}

int
nand_copyback_prog(device_t dev, uint32_t page, uint32_t col,
    void *buf, size_t len)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug("copyback prog page %x[%x]",  page, len);
	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (!can_write(nandbus))
		return (ENXIO);

	if (nandbus_send_command(nandbus, NAND_CMD_CHNG_WRITE_COL))
		return (ENXIO);

	if (nandbus_send_address(nandbus, row, col, -1))
		return (ENXIO);

	if (buf != NULL && len > 0)
		nandbus_write_buffer(nandbus, buf, len);

	if (send_end_program_page(nandbus, NAND_CMD_PROG_END))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_written++;

	return (0);
}

int
nand_copyback_prog_intlv(device_t dev, uint32_t page)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;

	debug("cache prog page %x", page);
	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (!can_write(nandbus))
		return (ENXIO);

	if (send_start_program_page(nandbus, row, 0))
		return (ENXIO);

	if (send_end_program_page(nandbus, NAND_CMD_PROG_INTLV))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_written++;

	return (0);
}

int
nand_prog_cache(device_t dev, uint32_t page, uint32_t col,
    void *buf, size_t len, uint8_t end)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;
	uint8_t command;

	debug("cache prog page %x[%x]",  page, len);
	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (!can_write(nandbus))
		return (ENXIO);

	if (send_start_program_page(nandbus, row, 0))
		return (ENXIO);

	nandbus_write_buffer(nandbus, buf, len);

	if (end)
		command = NAND_CMD_PROG_END;
	else
		command = NAND_CMD_PROG_CACHE;

	if (send_end_program_page(nandbus, command))
		return (ENXIO);

	DELAY(chip->t_prog);

	if (check_fail(nandbus))
		return (ENXIO);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_written++;

	return (0);
}

int
nand_read_cache(device_t dev, uint32_t page, uint32_t col,
    void *buf, size_t len, uint8_t end)
{
	struct nand_chip *chip;
	struct page_stat *pg_stat;
	device_t nandbus;
	uint32_t row;
	uint8_t command;

	debug("cache read page %x[%x] ", page, len);
	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nand_check_page_boundary(chip, page))
		return (ENXIO);

	page_to_row(&chip->chip_geom, page, &row);

	if (page != -1) {
		if (nandbus_send_command(nandbus, NAND_CMD_READ))
			return (ENXIO);

		if (nandbus_send_address(nandbus, row, col, -1))
			return (ENXIO);
	}

	if (end)
		command = NAND_CMD_READ_CACHE_END;
	else
		command = NAND_CMD_READ_CACHE;

	if (nandbus_send_command(nandbus, command))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	DELAY(chip->t_r);
	if (check_fail(nandbus))
		return (ENXIO);

	if (buf != NULL && len > 0)
		nandbus_read_buffer(nandbus, buf, len);

	pg_stat = &(chip->pg_stat[page]);
	pg_stat->page_raw_read++;

	return (0);
}

int
nand_get_feature(device_t dev, uint8_t feat, void *buf)
{
	struct nand_chip *chip;
	device_t nandbus;

	debug("nand get feature");

	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nandbus_send_command(nandbus, NAND_CMD_GET_FEATURE))
		return (ENXIO);

	if (nandbus_send_address(nandbus, -1, -1, feat))
		return (ENXIO);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	DELAY(chip->t_r);
	nandbus_read_buffer(nandbus, buf, 4);

	return (0);
}

int
nand_set_feature(device_t dev, uint8_t feat, void *buf)
{
	struct nand_chip *chip;
	device_t nandbus;

	debug("nand set feature");

	chip = device_get_softc(dev);
	nandbus = device_get_parent(dev);

	if (nandbus_send_command(nandbus, NAND_CMD_SET_FEATURE))
		return (ENXIO);

	if (nandbus_send_address(nandbus, -1, -1, feat))
		return (ENXIO);

	nandbus_write_buffer(nandbus, buf, 4);

	if (nandbus_start_command(nandbus))
		return (ENXIO);

	return (0);
}
#endif
