/*-
 * Copyright (C) 2009-2010 Semihalf, Grzegorz Bernacki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/socket.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/proc.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/condvar.h>

#include <dev/nand/nand.h>
#include <dev/nand/nandbus.h>
#include "nand_if.h"
#include "nfc_if.h"

#define DEBUG
#undef DEBUG
#ifdef DEBUG
#define debug(fmt, args...) do { \
    printf("NAND_BUS:" fmt "\n", ##args); } while (0)
#else
#define debug(fmt, args...)
#endif

#define NAND_NCS 4

static int nandbus_probe(device_t dev);
static int nandbus_attach(device_t dev);
static int nandbus_detach(device_t dev);

static int nandbus_child_location_str(device_t bus, device_t child, char *buf,
    size_t buflen);
static int nandbus_child_pnpinfo_str(device_t bus, device_t child, char *buf,
    size_t buflen);


static int nand_readid(device_t bus, uint8_t *man_id,
    uint8_t *dev_id);
static int nand_probe_onfi(device_t bus, uint8_t *onfi_compliant);
static int nand_reset(device_t bus);

struct nandbus_softc {
	device_t dev;
	struct cv nandbus_cv;
	struct mtx nandbus_mtx;
	uint8_t busy;
};

static device_method_t nandbus_methods[] = {
	/* device interface */
	DEVMETHOD(device_probe,		nandbus_probe),
	DEVMETHOD(device_attach,	nandbus_attach),
	DEVMETHOD(device_detach,	nandbus_detach),
	DEVMETHOD(device_shutdown,	bus_generic_shutdown),

	/* bus interface */
	DEVMETHOD(bus_print_child,	bus_generic_print_child),
	DEVMETHOD(bus_driver_added,	bus_generic_driver_added),
	DEVMETHOD(bus_child_pnpinfo_str, nandbus_child_pnpinfo_str),
	DEVMETHOD(bus_child_location_str, nandbus_child_location_str),

	{ 0, 0 }
};

devclass_t nandbus_devclass;

driver_t nandbus_driver = {
	"nandbus",
	nandbus_methods,
	sizeof(struct nandbus_softc)
};

DRIVER_MODULE(nandbus, nand, nandbus_driver, nandbus_devclass, 0, 0);

int
nandbus_create(device_t nfc)
{
	device_t child;

	child = device_add_child(nfc, "nandbus", -1);
	if (!child)
		return (ENODEV);

	bus_generic_attach(nfc);

	return(0);
}

void
nandbus_destroy(device_t nfc)
{
	device_t *children;
	int nchildren, i;

	mtx_lock(&Giant);
	/* detach & delete all children */
	if (!device_get_children(nfc, &children, &nchildren)) {
		for (i = 0; i < nchildren; i++)
			device_delete_child(nfc, children[i]);

		free(children, M_TEMP);
	}
	mtx_unlock(&Giant);
}

static int
nandbus_probe(device_t dev)
{
	device_set_desc(dev, "NAND bus");

	return (0);
}

static int
nandbus_attach(device_t dev)
{
	device_t child, nfc;
	struct nandbus_softc *sc;
	struct nandbus_ivar *ivar;
	struct nand_softc* nfc_sc;
	uint8_t cs, man_id, dev_id, onfi, found = 0;

	sc = device_get_softc(dev);
	sc->dev = dev;

	nfc = device_get_parent(dev);
	nfc_sc = device_get_softc(nfc);

	mtx_init(&sc->nandbus_mtx, "nandbus lock", MTX_DEF, 0);
	cv_init(&sc->nandbus_cv, "nandbus cv");

	/* Check each possible CS for existing nand devices */
	for (cs = 0; cs < NAND_NCS; cs++) {
		debug("probe chip select %x", cs);

		/* Select & reset chip */
		if (nandbus_select_cs(dev, cs))
			break;

		if (nand_reset(dev))
			continue;

		/* Read manufacturer and device id */
		if (nand_readid(dev, &man_id, &dev_id))
			continue;

		if (man_id == 0xff)
			continue;

		/* Check if chip is ONFI compliant */
		if (nand_probe_onfi(dev, &onfi) != 0) {
			continue;
		}

		ivar = malloc(sizeof(struct nandbus_ivar),
		    M_DEVBUF, M_NOWAIT);
		if (!ivar)
			return (ENOMEM);

		ivar->cs = cs;
		ivar->man_id = man_id;
		ivar->dev_id = dev_id;
		ivar->is_onfi = onfi;
		ivar->chip_cdev_name = nfc_sc->chip_cdev_name;

		child = device_add_child(dev, NULL, -1);
		device_set_ivars(child, ivar);
		found = 1;
	}

	bus_generic_attach(dev);
	return (0);
}

static int
nandbus_detach(device_t dev)
{
	struct nandbus_softc *sc;

	sc = device_get_softc(dev);

	mtx_destroy(&sc->nandbus_mtx);
	cv_destroy(&sc->nandbus_cv);

	bus_generic_detach(dev);

	return (0);
}

static int
nandbus_child_location_str(device_t bus, device_t child, char *buf,
    size_t buflen)
{
	struct nandbus_ivar *ivar = device_get_ivars(child);

	snprintf(buf, buflen, "at cs#%d", ivar->cs);
	return (0);
}

static int
nandbus_child_pnpinfo_str(device_t bus, device_t child, char *buf,
    size_t buflen)
{
	// XXX man id, model id ????
	*buf = '\0';
	return (0);
}

static int
nand_readid(device_t bus, uint8_t *man_id, uint8_t *dev_id)
{
	device_t nfc;

	if (!bus || !man_id || !dev_id)
		return (EINVAL);

	debug("read id");

	nfc = device_get_parent(bus);

	if (NFC_SEND_COMMAND(nfc, NAND_CMD_READ_ID)) {
		debug("Error : could not send READ ID command");
		return (ENXIO);
	}

	if (NFC_SEND_ADDRESS(nfc, -1, -1, 0)) {
		debug("Error : could not sent address to chip");
		return (ENXIO);
	}

	if (NFC_START_COMMAND(nfc) != 0) {
		debug("Error : could not start command");
		return (ENXIO);
	}

	DELAY(25);

	*man_id = NFC_READ_BYTE(nfc);
	*dev_id = NFC_READ_BYTE(nfc);

	debug("manufacturer id: %x chip id: %x", *man_id, *dev_id);

	return (0);
}

static int
nand_probe_onfi(device_t bus, uint8_t *onfi_compliant)
{
	device_t nfc;
	char onfi_id[] = {'o', 'n', 'f', 'i', '\0'};
	int i;

	debug("probing ONFI");

	nfc = device_get_parent(bus);

	if (NFC_SEND_COMMAND(nfc, NAND_CMD_READ_ID)) {
		debug("Error : could not sent READ ID command");
		return (ENXIO);
	}

	if (NFC_SEND_ADDRESS(nfc, -1, -1, ONFI_SIG_ADDR)) {
		debug("Error : could not sent address to chip");
		return (ENXIO);
	}

	if (NFC_START_COMMAND(nfc) != 0) {
		debug("Error : could not start command");
		return (ENXIO);
	}
	for (i = 0; onfi_id[i] != '\0'; i++)
		if (NFC_READ_BYTE(nfc) != onfi_id[i]) {
			debug("ONFI non-compliant");
			*onfi_compliant = 0;
			return (0);
		}

	debug("ONFI compliant");
	*onfi_compliant = 1;

	return (0);
}

static int
nand_reset(device_t bus)
{
	device_t nfc;
	debug("resetting...");

	nfc = device_get_parent(bus);

	if (NFC_SEND_COMMAND(nfc, NAND_CMD_RESET) != 0) {
		debug("Error : could not sent RESET command");
		return (ENXIO);
	}

	if (NFC_START_COMMAND(nfc) != 0) {
		debug("Error : could not start RESET command");
		return (ENXIO);
	}

	DELAY(1000);

	return (0);
}

void
nandbus_lock(device_t dev)
{
	struct nandbus_softc *sc;

	sc = device_get_softc(dev);

	mtx_lock(&sc->nandbus_mtx);
	if (sc->busy)
		cv_wait(&sc->nandbus_cv, &sc->nandbus_mtx);
	sc->busy = 1;
	mtx_unlock(&sc->nandbus_mtx);
}

void
nandbus_unlock(device_t dev)
{
	struct nandbus_softc *sc;

	sc = device_get_softc(dev);

	mtx_lock(&sc->nandbus_mtx);
	sc->busy = 0;
	cv_signal(&sc->nandbus_cv);
	mtx_unlock(&sc->nandbus_mtx);
}

int
nandbus_select_cs(device_t dev, uint8_t cs)
{

	return (NFC_SELECT_CS(device_get_parent(dev), cs));
}

int
nandbus_send_command(device_t dev, uint8_t command)
{
	int err;

	if ((err = NFC_SEND_COMMAND(device_get_parent(dev), command)))
		debug("Err: Could not send command %x, err %x", command, err);

	return (err);
}

int
nandbus_send_address(device_t dev, int32_t row, int32_t col, int8_t off)
{
	int err;

	if ((err = NFC_SEND_ADDRESS(device_get_parent(dev), row, col, off)))
		debug("Err: Could not send %x-%x-%x address, err %d",
		   row, col, off, err);

	return (err);
}

int
nandbus_start_command(device_t dev)
{
	int err;

	if ((err = NFC_START_COMMAND(device_get_parent(dev))))
		debug("Err: Could not start command, err %x", err);

	return (err);
}

void
nandbus_read_buffer(device_t dev, void *buf, uint32_t len)
{

	NFC_READ_BUF(device_get_parent(dev), buf, len);
}

void
nandbus_write_buffer(device_t dev, void *buf, uint32_t len)
{

	NFC_WRITE_BUF(device_get_parent(dev), buf, len);
}

int
nandbus_get_status(device_t dev, uint8_t *status)
{
	int err;

	if ((err = nandbus_send_command(dev, NAND_CMD_STATUS)))
		return (err);
	if ((err = nandbus_start_command(dev)))
		return (err);

	*status = NFC_READ_BYTE(device_get_parent(dev));

	return (0);
}

int
nandbus_wait_ready(device_t dev, uint8_t *status)
{
	uint32_t timeout;

	timeout = 1500;
	do {
		if (nandbus_get_status(dev, status))
			return (ENXIO);

		if (*status & NAND_STATUS_RDY)
			break;

		pause("nst", hz / 100);
		timeout--;
	} while (timeout);

	if (timeout)
		return (0);

	return (EBUSY);
}
