/*-
 * Copyright (C) 2009-2010 Semihalf, Grzegorz Bernacki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Integrated NAND controller driver */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/malloc.h>
#include <sys/rman.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/time.h>

#include <machine/bus.h>
#include <arm/mv/mvvar.h>
#include <arm/mv/mvwin.h>

#include <dev/nand/nand.h>
#include <dev/nand/nandbus.h>
#include "nfc_if.h"

#define NAND_DEBUG
#undef NAND_DEBUG
#ifdef NAND_DEBUG
#define debug(fmt, args...) do { \
	printf("NAND:" fmt "\n", ##args); } while (0)
#else
#define debug(fmt, args...)
#endif

#define MV_NAND_DATA	(MV_NAND_FLASH)
#define MV_NAND_COMMAND	(MV_NAND_FLASH + 1)
#define MV_NAND_ADDRESS	(MV_NAND_FLASH + 2)

#define MV_NAND_WRITE(addr, data)	bus_space_write_1(obio_tag, addr, 0, data);
#define MV_NAND_READ(addr)		bus_space_read_1(obio_tag, addr, 0);


struct mv_nand_softc {
	struct nand_softc nand_dev;
};

static int	mv_nand_attach(device_t dev);
static int	mv_nand_probe(device_t dev);
static int	mv_nand_send_command(device_t dev, uint8_t command);
static int	mv_nand_send_address(device_t dev, int32_t row,
    int32_t column, int8_t id);
static uint8_t	mv_nand_read_byte(device_t dev);
static void	mv_nand_read_buf(device_t dev, void* buf, uint32_t len);
static void	mv_nand_write_buf(device_t dev, void* buf, uint32_t len);
static int	mv_nand_select_cs(device_t dev, uint8_t cs);

static device_method_t mv_nand_methods[] = {
	DEVMETHOD(device_probe,		mv_nand_probe),
	DEVMETHOD(device_attach,	mv_nand_attach),

	DEVMETHOD(nfc_send_command,	mv_nand_send_command),
	DEVMETHOD(nfc_send_address,	mv_nand_send_address),
	DEVMETHOD(nfc_read_byte,	mv_nand_read_byte),
	DEVMETHOD(nfc_read_buf,		mv_nand_read_buf),
	DEVMETHOD(nfc_write_buf,	mv_nand_write_buf),
	DEVMETHOD(nfc_select_cs,	mv_nand_select_cs),

	{ 0, 0 },
};

static driver_t mv_nand_driver = {
	"nand",
	mv_nand_methods,
	sizeof(struct mv_nand_softc),
};

static devclass_t mv_nand_devclass;
DRIVER_MODULE(mv_nand, mbus, mv_nand_driver, mv_nand_devclass, 0, 0);

static int
mv_nand_probe(device_t dev)
{

	device_set_desc(dev, "Marvell NAND controller");
	return (BUS_PROBE_DEFAULT);
}

static int
mv_nand_attach(device_t dev)
{
	struct mv_nand_softc *sc;
	int err;

	sc = device_get_softc(dev);

	nand_init(&sc->nand_dev, dev, 0, NAND_ECC_SOFT,
	    NULL, NULL);

	err = nandbus_create(dev);

	return (err);
}

static int
mv_nand_send_command(device_t dev, uint8_t command)
{

	debug("mv_nand: send command %x", command);
	MV_NAND_WRITE(MV_NAND_COMMAND, command);

	return (0);
}


static int
mv_nand_send_address(device_t dev, int32_t row, int32_t column,
    int8_t id)
{
	struct mv_nand_softc *sc;
	sc = device_get_softc(dev);

	if (id != -1) {
		debug("mv_nand: send address %x", id);
		MV_NAND_WRITE(MV_NAND_ADDRESS, id);
	}

	/* TODO add number of address cycles to function parameters */
	if (column != -1) {
		debug("mv_nand: send address %x",
				column & 0xff);
		MV_NAND_WRITE(MV_NAND_ADDRESS, column & 0xff);
	}

	if (row != -1) {
		debug("mv_nand: send address %x %x",
				row & 0xff, (row >> 8) & 0xff);
		MV_NAND_WRITE(MV_NAND_ADDRESS, row & 0xff);
		MV_NAND_WRITE(MV_NAND_ADDRESS, (row >> 8) & 0xff);
	}

	return (0);
}

static uint8_t
mv_nand_read_byte(device_t dev)
{
	uint8_t data;

	data = MV_NAND_READ(MV_NAND_DATA);
	debug("mv_nand: read %x", data);

	return (data);
}

static void
mv_nand_read_buf(device_t dev, void* buf, uint32_t len)
{
	int i;
	uint8_t *b = (uint8_t*)buf;

	for (i = 0; i < len; i++) {
		b[i] = MV_NAND_READ(MV_NAND_DATA);
#ifdef NAND_DEBUG
		if (!(i % 16))
			printf("%s", i == 0 ? "mv_nand:\n" : "\n");
		printf(" %x", b[i]);
		if (i == len - 1)
			printf("\n");
#endif
	}

	return;
}

static void
mv_nand_write_buf(device_t dev, void* buf, uint32_t len)
{
	int i;
	uint8_t *b = (uint8_t*)buf;

	for (i = 0; i < len; i++) {
#ifdef NAND_DEBUG
		if (!(i % 16))
			printf("%s", i == 0 ? "mv_nand:\n" : "\n");
		printf(" %x", b[i]);
		if (i == len - 1)
			printf("\n");
#endif
		MV_NAND_WRITE(MV_NAND_DATA, b[i]);
	}

	return;
}

static int
mv_nand_select_cs(device_t dev, uint8_t cs)
{
	if (cs > 0)
		return (ENODEV);

	return (0);
}
